@@ -38,9 +38,11 @@ IMGT_HLA_URL = 'https://raw.githubusercontent.com/ANHIG/IMGTHLA/'
 # List of expression characters
 expression_chars = ['N', 'Q', 'L', 'S']
 
-ars_mapping_tables = ['dup_g', 'dup_lg', 'dup_lgx', 'g_group', 'lg_group', 'lgx_group']
+ars_mapping_tables = ['dup_g', 'dup_lg', 'dup_lgx', 'g_group', 'lg_group', 'lgx_group', 'exon_group', 'p_group']
 ARSMapping = namedtuple("ARSMapping", ars_mapping_tables)
 
+code_mapping_tables = ["alleles", "xx_codes", "who_alleles", "who_group", ]
+
 
 def get_n_field_allele(allele: str, n: int) -> str:
     """
@@ -54,7 +56,10 @@ def get_n_field_allele(allele: str, n: int) -> str:
     last_char = allele[-1]
     fields = allele.split(':')
     if last_char in expression_chars and len(fields) > n:
-        return ':'.join(fields[0:n]) + last_char
+
+        # don't actually do this;  it makes things like A*02:01:01L which is invalid
+        #return ':'.join(fields[0:n]) + last_char
+        return ':'.join(fields[0:n])
     else:
         return ':'.join(fields[0:n])
 
@@ -68,18 +73,21 @@ def get_2field_allele(a: str) -> str:
 
 
 def generate_ars_mapping(db_connection: sqlite3.Connection, imgt_version):
-    if db.tables_exists(db_connection, ars_mapping_tables):
+    if db.tables_exist(db_connection, ars_mapping_tables):
         dup_g = db.load_dict(db_connection, table_name='dup_g', columns=('allele', 'g_group'))
         dup_lg = db.load_dict(db_connection, table_name='dup_lg', columns=('allele', 'lg_group'))
         dup_lgx = db.load_dict(db_connection, table_name='dup_lgx', columns=('allele', 'lgx_group'))
         g_group = db.load_dict(db_connection, table_name='g_group', columns=('allele', 'g'))
         lg_group = db.load_dict(db_connection, table_name='lg_group', columns=('allele', 'lg'))
         lgx_group = db.load_dict(db_connection, table_name='lgx_group', columns=('allele', 'lgx'))
+        exon_group = db.load_dict(db_connection, table_name='exon_group', columns=('allele', 'exon'))
+        p_group = db.load_dict(db_connection, table_name='p_group', columns=('allele', 'p'))
         return ARSMapping(dup_g=dup_g, dup_lg=dup_lg, dup_lgx=dup_lgx,
-                          g_group=g_group, lg_group=lg_group, lgx_group=lgx_group)
+                          g_group=g_group, lg_group=lg_group,
+                          lgx_group=lgx_group, exon_group=exon_group, p_group=p_group)
 
-    ars_url = f'{IMGT_HLA_URL}{imgt_version}/wmda/hla_nom_g.txt'
-    df = pd.read_csv(ars_url, skiprows=6, names=["Locus", "A", "G"], sep=";").dropna()
+    ars_G_url = f'{IMGT_HLA_URL}{imgt_version}/wmda/hla_nom_g.txt'
+    df = pd.read_csv(ars_G_url, skiprows=6, names=["Locus", "A", "G"], sep=";").dropna()
 
     df['A'] = df['A'].apply(lambda a: a.split('/'))
     df = df.explode('A')
@@ -143,18 +151,35 @@ def generate_ars_mapping(db_connection: sqlite3.Connection, imgt_version):
     ])
     lgx_group = df_lgx.set_index('A')['lgx'].to_dict()
 
+    # exon
+    df_exon = pd.concat([df[['A', '3d']].rename(columns={'3d': 'exon'}), ])
+    exon_group = df_exon.set_index('A')['exon'].to_dict()
+
+    # P groups
+    ars_P_url = f'{IMGT_HLA_URL}{imgt_version}/wmda/hla_nom_p.txt'
+    df_P = pd.read_csv(ars_P_url, skiprows=6, names=["Locus", "A", "P"], sep=";").dropna()
+    df_P['A'] = df_P['A'].apply(lambda a: a.split('/'))
+    df_P      = df_P.explode('A')
+    df_P['A'] = df_P['Locus'] + df_P['A']
+    df_P['P'] = df_P['Locus'] + df_P['P']
+    p_group = df_P.set_index('A')['P'].to_dict()
+
+    # save
     db.save_dict(db_connection, table_name='dup_g', dictionary=dup_g, columns=('allele', 'g_group'))
     db.save_dict(db_connection, table_name='dup_lg', dictionary=dup_lg, columns=('allele', 'lg_group'))
     db.save_dict(db_connection, table_name='dup_lgx', dictionary=dup_lgx, columns=('allele', 'lgx_group'))
     db.save_dict(db_connection, table_name='g_group', dictionary=g_group, columns=('allele', 'g'))
     db.save_dict(db_connection, table_name='lg_group', dictionary=lg_group, columns=('allele', 'lg'))
     db.save_dict(db_connection, table_name='lgx_group', dictionary=lgx_group, columns=('allele', 'lgx'))
+    db.save_dict(db_connection, table_name='exon_group', dictionary=exon_group, columns=('allele', 'exon'))
+    db.save_dict(db_connection, table_name='p_group', dictionary=exon_group, columns=('allele', 'p'))
 
     return ARSMapping(dup_g=dup_g, dup_lg=dup_lg, dup_lgx=dup_lgx,
-                      g_group=g_group, lg_group=lg_group, lgx_group=lgx_group)
+                      g_group=g_group, lg_group=lg_group,
+                      lgx_group=lgx_group, exon_group=exon_group, p_group=p_group)
 
 
-def generate_alleles_and_xx_codes(db_connection: sqlite3.Connection, imgt_version):
+def generate_alleles_and_xx_codes_and_who(db_connection: sqlite3.Connection, imgt_version, ars_mappings):
     """
     Checks to see if there's already an allele list file for the `imgt_version`
     in the `data_dir` directory. If not, will download the file and create
@@ -181,15 +206,22 @@ def generate_alleles_and_xx_codes(db_connection: sqlite3.Connection, imgt_versio
 
     :param db_connection: Database connection to the sqlite database
     :param imgt_version: IMGT database version
+    :param ars_mappings: ARSMapping object to ARS mapping tables
     :return: None, updates self
     """
-
-    if db.table_exists(db_connection, 'alleles'):
+    if db.tables_exist(db_connection, code_mapping_tables):
         valid_alleles = db.load_set(db_connection, 'alleles')
+
+        who_alleles = db.load_set(db_connection, 'who_alleles')
+        who_group = db.load_dict(db_connection, 'who_group',
+                                 ('who', 'allele_list'))
+        who_group = {k: v.split('/') for k, v in who_group.items()}
+
         xx_codes = db.load_dict(db_connection, 'xx_codes',
                                 ('allele_1d', 'allele_list'))
         xx_codes = {k: v.split('/') for k, v in xx_codes.items()}
-        return valid_alleles, xx_codes
+
+        return valid_alleles, who_alleles, xx_codes, who_group
 
     # Create a Pandas DataFrame from the mac_code list file
     # Skip the header (first 6 lines) and use only the Allele column
@@ -203,6 +235,8 @@ def generate_alleles_and_xx_codes(db_connection: sqlite3.Connection, imgt_versio
     # All 2-field, 3-field and the original Alleles are considered valid alleles
     allele_df['2d'] = allele_df['Allele'].apply(get_2field_allele)
     allele_df['3d'] = allele_df['Allele'].apply(get_3field_allele)
+    # this says all 3rd and 2nd field versions of longer alleles are valid
+    who_alleles = set(allele_df['Allele'])
     valid_alleles = set(allele_df['Allele']). \
         union(set(allele_df['2d'])). \
         union(set(allele_df['3d']))
@@ -226,13 +260,53 @@ def generate_alleles_and_xx_codes(db_connection: sqlite3.Connection, imgt_versio
 
     # Save this version of the valid alleles
     db.save_set(db_connection, 'alleles', valid_alleles, 'allele')
+    # Save this version of the who alleles
+    db.save_set(db_connection, 'who_alleles', who_alleles, 'allele')
     # Save this version of xx codes
     flat_xx_codes = {k: '/'.join(sorted(v, key=functools.cmp_to_key(smart_sort_comparator)))
                      for k, v in xx_codes.items()}
     db.save_dict(db_connection, 'xx_codes', flat_xx_codes,
                  ('allele_1d', 'allele_list'))
 
-    return valid_alleles, xx_codes
+    # W H O
+    # Create WHO mapping from the unique alleles in the 1-field column
+    unique_alleles = allele_df['Allele'].unique()
+    who_df1 = pd.DataFrame(unique_alleles, columns=['Allele'])
+    who_df1['nd'] = allele_df['Allele'].apply(lambda x: x.split(":")[0])
+    # Create WHO mapping from the unique alleles in the 2-field column
+    who_df2 = pd.DataFrame(unique_alleles, columns=['Allele'])
+    who_df2['nd'] = allele_df['Allele'].apply(get_2field_allele)
+    # Create WHO mapping from the unique alleles in the 3-field column
+    who_df3 = pd.DataFrame(unique_alleles, columns=['Allele'])
+    who_df3['nd'] = allele_df['Allele'].apply(get_3field_allele)
+    # Combine n-field dataframes in 1
+
+    # Create g_codes expansion mapping from the same tables used to reduce to G
+    g_df = pd.DataFrame(list(ars_mappings.g_group.items()),columns = ['Allele','nd']) 
+
+    # Create p_codes expansion mapping from the p_group table
+    p_df = pd.DataFrame(list(ars_mappings.p_group.items()),columns = ['Allele','nd']) 
+
+    who_codes = pd.concat([who_df1, who_df2, who_df3, g_df, p_df])
+
+    # remove valid alleles from who_codes to avoid recursion
+    for k in who_alleles:
+        if k in who_codes['nd']:
+            who_codes.drop(labels=k, axis='index')
+
+    # drop duplicates
+    who_codes = who_codes.drop_duplicates()
+
+    # who_codes maps a first field name to its 2 field expansion
+    who_group = who_codes.groupby(['nd']).apply(lambda x: list(x['Allele'])).to_dict()
+
+    # dictionary
+    flat_who_group = {k: '/'.join(sorted(v, key=functools.cmp_to_key(smart_sort_comparator)))
+                      for k, v in who_group.items()}
+    db.save_dict(db_connection, 'who_group', flat_who_group,
+                 columns=('who', 'allele_list'))
+
+    return valid_alleles, who_alleles, xx_codes, who_group
 
 
 def generate_mac_codes(db_connection: sqlite3.Connection, refresh_mac: bool):
@@ -307,7 +381,7 @@ def to_serological_name(locus_name: str):
     """
     locus, sero_number = locus_name.split('*')
     sero_locus = locus[:2]
-    if (sero_locus == "C"):
+    if sero_locus == "C":
         sero_locus = "Cw"
     sero_name = sero_locus + sero_number
     return sero_name
@@ -378,8 +452,8 @@ def generate_v2_to_v3_mapping(db_connection: sqlite3.Connection, imgt_version):
         #  deleted alleles as reference.
         # Temporary Example
         v2_to_v3_example = {
-            "A*0104": "A*01:04N",
-            "A*0105N": "A*01:04N",
+            "A*0104": "A*01:04:01:01N",
+            "A*0105N": "A*01:04:01:01N",
             "A*0111": "A*01:11N",
             "A*01123": "A*01:123N",
             "A*0115": "A*01:15N",
