@@ -1,7 +1,8 @@
 # -*- coding: utf-8 -*-
 #
 #    py-ard
-#    Copyright (c) 2020 Be The Match operated by National Marrow Donor Program. All Rights Reserved.
+#    Copyright (c) 2020 Be The Match operated by National Marrow Donor Program.
+#    All Rights Reserved.
 #
 #    This library is free software; you can redistribute it and/or modify it
 #    under the terms of the GNU Lesser General Public License as published
@@ -21,19 +22,34 @@
 #    > http://www.opensource.org/licenses/lgpl-license.php
 #
 import functools
-import gc
+import sys
 import re
 from typing import Iterable
 
 from . import db
 from .data_repository import generate_ars_mapping, \
-    generate_mac_codes, generate_alleles_and_xx_codes, \
+    generate_mac_codes, generate_alleles_and_xx_codes_and_who, \
     generate_serology_mapping, generate_v2_to_v3_mapping
 from .db import is_valid_mac_code, mac_code_to_alleles, v2_to_v3_allele
 from .smart_sort import smart_sort_comparator
 
 HLA_regex = re.compile("^HLA-")
 
+# Set the lru cache size with
+# >>> import pyard
+# >>> pyard.max_cache_size = 1_000_000
+# >>> ard = pyard.ARD()
+
+max_cache_size = 1_000
+default_config = {
+    "reduce_serology": True,
+    "reduce_v2": True,
+    "reduce_3field": True,
+    "reduce_P": True,
+    "reduce_XX": True,
+    "reduce_MAC": True,
+    "map_drb345_to_drbx": True,
+    "verbose_log": True}
 
 class ARD(object):
     """
@@ -41,30 +57,36 @@ class ARD(object):
     Allows reducing alleles and allele code(MAC) to G, lg and lgx levels.
     """
 
-    def __init__(self, imgt_version: str = 'Latest',
-                 remove_invalid: bool = True,
-                 data_dir: str = None,
-                 refresh_mac: bool = False) -> None:
+    def __init__(self, imgt_version: str = 'Latest', data_dir : str = None, config: dict = None):
+
         """
         ARD will load valid alleles, xx codes and MAC mappings for the given
         version of IMGT database, downloading and generating the database if
         not already present.
 
         :param imgt_version: IMGT HLA database version
-        :param remove_invalid: report only valid alleles for this version
         :param data_dir: directory path to store cached data
+        :param config: directory of configuration options
         """
-        self._remove_invalid = remove_invalid
+        self._data_dir = data_dir
+        self._config = default_config.copy()
+        if config:
+            self._config.update(config)
 
         # Create a database connection for writing
         self.db_connection = db.create_db_connection(data_dir, imgt_version)
 
         # Load MAC codes
-        generate_mac_codes(self.db_connection, refresh_mac)
-        # Load Alleles and XX Codes
-        self.valid_alleles, self.xx_codes = generate_alleles_and_xx_codes(self.db_connection, imgt_version)
+        generate_mac_codes(self.db_connection, False)
         # Load ARS mappings
         self.ars_mappings = generate_ars_mapping(self.db_connection, imgt_version)
+        # Load Alleles and XX Codes
+        self.valid_alleles, self.who_alleles, self.xx_codes, self.who_group = \
+            generate_alleles_and_xx_codes_and_who(self.db_connection, imgt_version, self.ars_mappings )
+        # Load Alleles and XX Codes
+        self.valid_alleles, self.who_alleles, self.xx_codes, self.who_group = generate_alleles_and_xx_codes_and_who(self.db_connection, imgt_version, self.ars_mappings)
+
+        
         # Load Serology mappings
         generate_serology_mapping(self.db_connection, imgt_version)
         # Load V2 to V3 mappings
@@ -74,7 +96,10 @@ class ARD(object):
         self.db_connection.close()
 
         # reference data is read-only and can be frozen
-        gc.freeze()
+        # Works only for Python >= 3.9
+        if sys.version_info.major == 3 and sys.version_info.minor >= 9:
+            import gc
+            gc.freeze()
 
         # Re-open the connection in read-only mode as we're not updating it anymore
         self.db_connection = db.create_db_connection(data_dir, imgt_version, ro=True)
@@ -86,7 +111,7 @@ class ARD(object):
         """
         self.db_connection.close()
 
-    @functools.lru_cache(maxsize=1000)
+    @functools.lru_cache(maxsize=max_cache_size)
     def redux(self, allele: str, ars_type: str) -> str:
         """
         Does ARS reduction with allele and ARS type
@@ -98,7 +123,6 @@ class ARD(object):
         :return: ARS reduced allele
         :rtype: str
         """
-
         # deal with leading 'HLA-'
         if HLA_regex.search(allele):
             hla, allele_name = allele.split("-")
@@ -108,10 +132,15 @@ class ARD(object):
             else:
                 return redux_allele
 
-        # Alleles ending with P or G are valid_alleles
-        if allele.endswith(('P', 'G')):
-            allele = allele[:-1]
+        # g_group maps alleles to their g_group
+        # note: this includes mappings for shortened version of alleles 
+        # C*12:02:02:01 => C*12:02:01G 
+        # C*12:02:02    => C*12:02:01G 
+        # C*12:02       => C*12:02:01G
 
+        if allele.endswith(('P', 'G')):
+            if ars_type in ["lg", "lgx", "G"]:
+                 allele = allele[:-1]
         if ars_type == "G" and allele in self.ars_mappings.g_group:
             if allele in self.ars_mappings.dup_g:
                 return self.ars_mappings.dup_g[allele]
@@ -135,16 +164,30 @@ class ARD(object):
                 # for 'lgx' when allele is not in G group,
                 # return allele with only first 2 field
                 return ':'.join(allele.split(':')[0:2])
-        else:
-            if self._remove_invalid:
-                if self._is_valid_allele(allele):
-                    return allele
-                else:
-                    return ''
+        elif ars_type == "W":
+            # new ars_type which is full WHO expansion
+            if self._is_who_allele(allele):
+                return allele
+            if allele in self.who_group:
+                return self.redux_gl("/".join(self.who_group[allele]), ars_type)
             else:
                 return allele
+        elif ars_type == "exon":
+            if allele in self.ars_mappings.exon_group:
+                return self.ars_mappings.exon_group[allele]
+            else:
+                # for 'exon' return allele with only first 3 fields
+                return ':'.join(allele.split(':')[0:3])
+        else:
+            if allele.endswith(('P', 'G')):
+                 allele = allele[:-1]
+            if self._is_valid_allele(allele):
+                return allele
+            else:
+                # TODO: raise error
+                return ''
 
-    @functools.lru_cache(maxsize=1000)
+    @functools.lru_cache(maxsize=max_cache_size)
     def redux_gl(self, glstring: str, redux_type: str) -> str:
         """
         Does ARS reduction with gl string and ARS type
@@ -180,24 +223,28 @@ class ARD(object):
                                    key=functools.cmp_to_key(smart_sort_comparator)))
 
         # Handle V2 to V3 mapping
-        if self.is_v2(glstring):
+        if self._config["reduce_v2"] and self.is_v2(glstring):
             glstring = self._map_v2_to_v3(glstring)
             return self.redux_gl(glstring, redux_type)
 
         # Handle Serology
-        if self.is_serology(glstring):
+        if self._config["reduce_serology"] and self.is_serology(glstring):
             alleles = self._get_alleles_from_serology(glstring)
             return self.redux_gl("/".join(alleles), redux_type)
 
-        loc_allele = glstring.split(":")
-        loc_antigen, code = loc_allele[0], loc_allele[1]
+        if ":" in glstring: 
+            loc_allele = glstring.split(":")
+            loc_antigen, code = loc_allele[0], loc_allele[1]
+        else:
+            # TODO: raise error
+            return ''
 
         # Handle XX codes
-        if self.is_XX(glstring, loc_antigen, code):
+        if self._config["reduce_XX"] and self.is_XX(glstring, loc_antigen, code):
             return self.redux_gl("/".join(self.xx_codes[loc_antigen]), redux_type)
 
         # Handle MAC
-        if self.is_mac(glstring):
+        if self._config["reduce_MAC"] and self.is_mac(glstring):
             if is_valid_mac_code(self.db_connection, code):
                 if HLA_regex.search(glstring):
                     # Remove HLA- prefix
@@ -224,7 +271,7 @@ class ARD(object):
         """
         A serology has the locus name (first 2 letters for DRB1, DQB1)
         of the allele followed by numerical antigen.
-        Cw is the serlogical designation for HLA-C 
+        Cw is the serological designation for HLA-C
 
         :param allele: The allele to test for serology
         :return: True if serology
@@ -246,9 +293,11 @@ class ARD(object):
 
     @staticmethod
     def is_mac(gl: str) -> bool:
+        # TODO: need a more stringent test here
+	# not all strings are MACs e.g. ":THISISNOTAMAC"
         """
-        MAC has there are non-digit characters after the : character,
-        then it's a MAC.
+        MAC has non-digit characters after the : character.
+
         :param gl: glstring to test if it has a MAC code
         :return: True if MAC
         """
@@ -256,14 +305,27 @@ class ARD(object):
 
     @staticmethod
     def is_v2(allele: str) -> bool:
+        # TODO: need a more stringent test here
+        # not all strings with "*" but not ":" are v2 nomenclature e.g. "this s*it"
         """
         Version 2 of the nomenclature is a single field.
         It does not have any ':' field separator.
-        Eg: A*0104
+            Eg: A*0104
+        Exceptions:
+            DRB3*NNNN is not v2 allele
         :param allele: Possible allele
         :return: Is the allele in V2 nomenclature
         """
-        return '*' in allele and ':' not in allele
+        return '*' in allele and ':' not in allele \
+               and not allele.endswith('*NNNN')
+
+    def _is_who_allele(self, allele):
+        """
+        Test if allele is a WHO allele in the current imgt database
+        :param allele: Allele to test
+        :return: bool to indicate if allele is valid
+        """
+        return allele in self.who_alleles
 
     def _is_valid_allele(self, allele):
         """
@@ -271,9 +333,7 @@ class ARD(object):
         :param allele: Allele to test
         :return: bool to indicate if allele is valid
         """
-        if self._remove_invalid:
-            return allele in self.valid_alleles
-        return True
+        return allele in self.valid_alleles
 
     def _get_alleles(self, code, locus_antigen) -> Iterable[str]:
         """
@@ -293,17 +353,11 @@ class ARD(object):
         else:
             alleles = [f'{locus_antigen}:{a}' for a in alleles]
 
-        if self._remove_invalid:
-            return filter(self._is_valid_allele, alleles)
-        else:
-            return alleles
+        return filter(self._is_valid_allele, alleles)
 
     def _get_alleles_from_serology(self, serology) -> Iterable[str]:
         alleles = db.serology_to_alleles(self.db_connection, serology)
-        if self._remove_invalid:
-            return filter(self._is_valid_allele, alleles)
-        else:
-            return alleles
+        return filter(self._is_valid_allele, alleles)
 
     @staticmethod
     def _combine_with_colon(digits_field):
@@ -363,8 +417,19 @@ class ARD(object):
         :return: allele or empty
         :rtype: bool
         """
-        if allele == '':
+        if allele == '' or allele.endswith('*'):
             return False
+
+        # validate allele without the 'HLA-' prefix
+        if HLA_regex.search(allele):
+            # remove 'HLA-' prefix
+            allele = allele[4:]
+
+        if '*' in allele:
+            alphanum_allele = allele.replace('*', '').replace(':', '')
+            if not alphanum_allele.isalnum():
+                return False
+
         if not self.is_mac(allele) and \
                 not self.is_serology(allele) and \
                 not self.is_v2(allele):
@@ -372,10 +437,11 @@ class ARD(object):
             if allele.endswith(('P', 'G')):
                 # remove the last character
                 allele = allele[:-1]
-            # validate allele without the 'HLA-' prefix
-            if HLA_regex.search(allele):
-                # remove 'HLA-' prefix
-                allele = allele[4:]
+                if self._is_valid_allele(allele):
+                    return True
+                else:
+                    # reduce to 2 field for things like DPB1*28:01:01G 
+                    allele = ':'.join(allele.split(':')[0:2])
             return self._is_valid_allele(allele)
         return True
 
@@ -471,3 +537,10 @@ class ARD(object):
         if self.is_v2(v2_allele):
             return self._map_v2_to_v3(v2_allele)
         return v2_allele
+
+    def refresh_mac_codes(self) -> None:
+        """
+        Refreshes MAC code for the current IMGT db version.
+        :return: None
+        """
+        generate_mac_codes(self.db_connection, True)
